from io import BytesIO
from typing import Iterator, List, TypeVar

import PIL
from pymemri.data.schema import EmailMessage, Item, Message, Photo, Tweet
from pymemri.pod.client import PodClient


def item_to_data(item: Item, client: PodClient):
    if isinstance(item, Photo):
        return photo_to_PIL(item, client)
    elif isinstance(item, EmailMessage) or isinstance(item, Message):
        return item.content
    elif isinstance(item, Tweet):
        return item.message
    else:
        raise NotImplementedError(
            f"Converting Item of type {type(item)} is not implemented."
        )


def photo_to_PIL(photo: Photo, client: PodClient):
    photo = client.get_photo(photo.id)
    return PIL.Image.open(BytesIO(photo.data))


T = TypeVar("T")

def batch_list(list_: List[T], batch_size: int) -> Iterator[List[T]]:
    """Returns a generator from `list_` that yields batches of size `batch_size`

    Args:
        list_ (List[T]): List to batch
        batch_size (int): Size of batches. The final batch might be smaller.

    Yields:
        Generator[List[T], None, None]: Generator of batches.
    """
    for idx in range(0, len(list_), batch_size):
        yield list_[idx:idx + batch_size]
